@@ -5,16 +5,17 @@ import os
 import socket
 import ssl
 import stat
-import sys
 import types
 import warnings
 from dataclasses import dataclass
 from ssl import SSLContext, VerifyFlags, VerifyMode  # type: ignore
-from typing import Any, AnyStr, Dict, List, Mapping, Optional, Type, Union
+from time import time
+from typing import Any, AnyStr, Dict, List, Mapping, Optional, Tuple, Type, Union
+from wsgiref.handlers import format_date_time
 
-import pytoml
+import toml
 
-from .logging import AccessLogger
+from .logging import Logger
 
 BYTES = 1
 OCTETS = 1
@@ -27,38 +28,47 @@ FilePath = Union[AnyStr, os.PathLike]
 class Sockets:
     secure_sockets: List[socket.socket]
     insecure_sockets: List[socket.socket]
+    quic_sockets: List[socket.socket]
 
 
 class Config:
-
-    _access_logger: Optional[AccessLogger] = None
-    _error_log_target: Optional[str] = None
     _bind = ["127.0.0.1:8000"]
     _insecure_bind: List[str] = []
+    _quic_bind: List[str] = []
+    _log: Optional[Logger] = None
 
     access_log_format = "%(h)s %(S)s %(r)s %(s)s %(b)s %(D)s"
-    access_log_target: Optional[str] = None
-    access_logger_class = AccessLogger
-    alpn_protocols = ["h2", "http/1.1"]
+    accesslog: Union[logging.Logger, str, None] = None
+    alpn_protocols = ["h3-22", "h2", "http/1.1"]
     application_path: str
     backlog = 100
     ca_certs: Optional[str] = None
     certfile: Optional[str] = None
     ciphers: str = "ECDHE+AESGCM"
     debug = False
-    error_logger: Optional[logging.Logger] = None
+    dogstatsd_tags = ""
+    errorlog: Union[logging.Logger, str, None] = "-"
+    group: Optional[int] = None
     h11_max_incomplete_size = 16 * 1024 * BYTES
     h2_max_concurrent_streams = 100
     h2_max_header_list_size = 2 ** 16
     h2_max_inbound_frame_size = 2 ** 14 * OCTETS
+    include_server_header = True
     keep_alive_timeout = 5 * SECONDS
     keyfile: Optional[str] = None
+    logger_class = Logger
+    loglevel: str = "info"
+    max_app_queue_size: int = 10
     pid_path: Optional[str] = None
     root_path = ""
     shutdown_timeout = 60 * SECONDS
     ssl_handshake_timeout = 60 * SECONDS
     startup_timeout = 60 * SECONDS
+    statsd_host: Optional[str] = None
+    statsd_prefix = ""
+    umask: Optional[int] = None
     use_reloader = False
+    user: Optional[int] = None
     verify_flags: Optional[VerifyFlags] = None
     verify_mode: Optional[VerifyMode] = None
     websocket_max_message_size = 16 * 1024 * 1024 * BYTES
@@ -72,31 +82,10 @@ class Config:
     cert_reqs = property(None, set_cert_reqs)
 
     @property
-    def access_logger(self) -> AccessLogger:
-        if self._access_logger is None:
-            self._access_logger = self.access_logger_class(
-                self.access_log_format, self.access_log_target
-            )
-        return self._access_logger
-
-    @access_logger.setter
-    def access_logger(self, value: logging.Logger) -> None:
-        self._access_logger = self.access_logger_class(self.access_log_format, value)
-
-    @property
-    def error_log_target(self) -> Optional[str]:
-        return self._error_log_target
-
-    @error_log_target.setter
-    def error_log_target(self, value: Optional[str]) -> None:
-        self._error_log_target = value
-        if self.error_log_target is not None:
-            self.error_logger = logging.getLogger("hypercorn.error")
-            if self.error_log_target == "-":
-                self.error_logger.addHandler(logging.StreamHandler(sys.stderr))
-            else:
-                self.error_logger.addHandler(logging.FileHandler(self.error_log_target))
-            self.error_logger.setLevel(logging.INFO)
+    def log(self) -> Logger:
+        if self._log is None:
+            self._log = self.logger_class(self)
+        return self._log
 
     @property
     def bind(self) -> List[str]:
@@ -120,6 +109,17 @@ class Config:
         else:
             self._insecure_bind = value
 
+    @property
+    def quic_bind(self) -> List[str]:
+        return self._quic_bind
+
+    @quic_bind.setter
+    def quic_bind(self, value: Union[List[str], str]) -> None:
+        if isinstance(value, str):
+            self._quic_bind = [value]
+        else:
+            self._quic_bind = value
+
     def create_ssl_context(self) -> Optional[SSLContext]:
         if not self.ssl_enabled:
             return None
@@ -133,12 +133,10 @@ class Config:
         context.options |= cipher_opts  # RFC 7540 Section 9.2: MUST be TLS >=1.2
         context.options |= ssl.OP_NO_COMPRESSION  # RFC 7540 Section 9.2.1: MUST disable compression
         context.set_alpn_protocols(self.alpn_protocols)
-        try:
-            context.set_npn_protocols(self.alpn_protocols)
-        except NotImplementedError:
-            pass  # NPN is not necessarily available
 
-        context.load_cert_chain(certfile=self.certfile, keyfile=self.keyfile)
+        if self.certfile is not None and self.keyfile is not None:
+            context.load_cert_chain(certfile=self.certfile, keyfile=self.keyfile)
+
         if self.ca_certs is not None:
             context.load_verify_locations(self.ca_certs)
         if self.verify_mode is not None:
@@ -156,17 +154,21 @@ class Config:
         if self.ssl_enabled:
             secure_sockets = self._create_sockets(self.bind)
             insecure_sockets = self._create_sockets(self.insecure_bind)
+            quic_sockets = self._create_sockets(self.quic_bind, socket.SOCK_DGRAM)
         else:
             secure_sockets = []
             insecure_sockets = self._create_sockets(self.bind)
-        return Sockets(secure_sockets, insecure_sockets)
+            quic_sockets = []
+        return Sockets(secure_sockets, insecure_sockets, quic_sockets)
 
-    def _create_sockets(self, binds: List[str]) -> List[socket.socket]:
+    def _create_sockets(
+        self, binds: List[str], type_: int = socket.SOCK_STREAM
+    ) -> List[socket.socket]:
         sockets: List[socket.socket] = []
         for bind in binds:
             binding: Any = None
             if bind.startswith("unix:"):
-                sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
+                sock = socket.socket(socket.AF_UNIX, type_)
                 binding = bind[5:]
                 try:
                     if stat.S_ISSOCK(os.stat(binding).st_mode):
@@ -174,34 +176,59 @@ class Config:
                 except FileNotFoundError:
                     pass
             elif bind.startswith("fd://"):
-                sock = socket.fromfd(int(bind[5:]), socket.AF_UNIX, socket.SOCK_STREAM)
+                sock = socket.fromfd(int(bind[5:]), socket.AF_UNIX, type_)
             else:
+                bind = bind.replace("[", "").replace("]", "")
                 try:
                     value = bind.rsplit(":", 1)
                     host, port = value[0], int(value[1])
                 except (ValueError, IndexError):
                     host, port = bind, 8000
-                sock = socket.socket(
-                    socket.AF_INET6 if ":" in host else socket.AF_INET, socket.SOCK_STREAM
-                )
+                sock = socket.socket(socket.AF_INET6 if ":" in host else socket.AF_INET, type_)
                 if self.workers > 1:
                     try:
-                        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)  # type: ignore
+                        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
                     except AttributeError:
                         pass
                 binding = (host, port)
 
             sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
-            if binding is not None:
+
+            if bind.startswith("unix:"):
+                if self.umask is not None:
+                    current_umask = os.umask(self.umask)
                 sock.bind(binding)
+                if self.user is not None and self.group is not None:
+                    os.chown(binding, self.user, self.group)
+                if self.umask is not None:
+                    os.umask(current_umask)
+            elif bind.startswith("fd://"):
+                pass
+            else:
+                sock.bind(binding)
+
             sock.setblocking(False)
             try:
-                sock.set_inheritable(True)  # type: ignore
+                sock.set_inheritable(True)
             except AttributeError:
                 pass
             sockets.append(sock)
         return sockets
 
+    def response_headers(self, protocol: str) -> List[Tuple[bytes, bytes]]:
+        headers = [(b"date", format_date_time(time()).encode("ascii"))]
+        if self.include_server_header:
+            headers.append((b"server", f"hypercorn-{protocol}".encode("ascii")))
+        for bind in self._quic_bind:
+            port = int(bind.split(":")[-1])
+            headers.append((b"alt-svc", b'quic=":%d"; ma=2592000;' % port))
+
+        return headers
+
+    def set_statsd_logger_class(self, statsd_logger: Type[Logger]) -> None:
+        if self.logger_class == Logger and self.statsd_host is not None:
+            self.logger_class = statsd_logger
+
     @classmethod
     def from_mapping(
         cls: Type["Config"], mapping: Optional[Mapping[str, Any]] = None, **kwargs: Any
@@ -215,7 +242,7 @@ class Config:
 
             config = {'keep_alive_timeout': 10}
             Config.from_mapping(config)
-            Config.form_mapping(keep_alive_timeout=10)
+            Config.from_mapping(keep_alive_timeout=10)
 
         Arguments:
             mapping: Optionally a mapping object.
@@ -247,7 +274,7 @@ class Config:
             filename: The filename which gives the path to the file.
         """
         file_path = os.fspath(filename)
-        spec = importlib.util.spec_from_file_location("module.name", file_path)  # type: ignore
+        spec = importlib.util.spec_from_file_location("module.name", file_path)
         module = importlib.util.module_from_spec(spec)
         spec.loader.exec_module(module)  # type: ignore
         return cls.from_object(module)
@@ -267,7 +294,7 @@ class Config:
         """
         file_path = os.fspath(filename)
         with open(file_path) as file_:
-            data = pytoml.load(file_)
+            data = toml.load(file_)
         return cls.from_mapping(data)
 
     @classmethod
